#!/usr/bin/env python
# -*- coding: utf-8 -*-

import FftUtils
from WindowType import WindowType
from WindowAttributes import WindowAttributes
from AudioSignal import AudioSignal
from Duet import Duet
from Nmf import Nmf, DistanceType
from Repet import Repet, RepetType


__version__ = '0.1.3'

__title__ = 'nussl'
__description__ = 'A flexible sound source separation library.'
__uri__ = 'https://github.com/interactiveaudiolab/nussl'

__author__ = 'C. Grief, E. Manilow, F. Pishdadian'
__email__ = 'ethanmanilow2015@u.northwestern.edu'

__license__ = 'Apache 2.0'
__copyright__ = 'Copyright (c) 2016 Interactive Audio Lab'
